#+xcvb (module (:build-depends-on () :depends-on ("/asdf")))
#| ;;; cl-launch 4.0.4 lisp header

|# ;;;; Silence our lisp implementation for quiet batch use...

#| We'd like to evaluate as little as possible of the code without compilation.
 This poses a typical bootstrapping problem: the more sophistication we want
 to distinguish what to put where in what dynamic environment, the more code
 we have to evaluate before we may actually load compiled files. And, then,
 it is a waste of time to try to compile said code into a file. Moving things
 to the shell can only help so much, and reduces flexibility. Our best bet is
 to tell sbcl or cmucl to not try to optimize too hard.
|#
(eval-when (:compile-toplevel :load-toplevel :execute)
  ;; Configure the printer
  (setf *print-readably* nil ; allegro 5.0 may bork without this
        *print-level* nil)
  ;; Hush the compiler and loader.
  (setf *load-verbose* nil *compile-verbose* nil *compile-print* nil *load-print* nil)
  ;; The code below exemplifies how to try super-hard to load ASDF 3 from standard locations,
  ;; by trying nice methods first, then increasingly desperate ones.
  ;; Stage 1 is to load ASDF at all.
  ;; Stage 2 is to upgrade to whichever ASDF installation the user has configured (if any).
  ;; Versions older than ASDF 3.1 need to be told about ~/common-lisp/
  ;; ASDF 1 has no output translation layer, so can be configured to load ASDF 3
  ;; only if ASDF 3 is in a predictable place under the user's homedir, thus
  ;; ~/common-lisp/asdf/ or ~/.local/share/common-lisp/source/asdf/ only.
  (block nil
    (let ((required-asdf-version  "3.0.1")
          (verbose *load-verbose*))
      (labels ((asdf-symbol (name)
                 (and (find-package :asdf) (find-symbol (string name) :asdf)))
               (asdf-call (name &rest args)
                 (apply (asdf-symbol name) args))
               (asdf-version ()
                 (when (find-package :asdf)
                   (or (symbol-value (or (asdf-symbol '*asdf-version*)
                                         (asdf-symbol '*asdf-revision*)))
                       "1.0")))
               (maybe-display (message)
                 (when (and verbose message) (format t "~&~A~%" message)))
               (call-maybe-verbosely (message function &rest args)
                 (cond
                   (verbose
                    (maybe-display message)
                    (apply function args))
                   (t
                    #+abcl ;; Bug in ABCL 1.3.0: without this, loading asdf.fasl shows warnings
                    (let* ((uc (asdf-symbol '*uninteresting-conditions*))
                           (vars (when uc (list uc)))
                           (vals (when uc (list (cons 'warning (symbol-value uc))))))
                      (progv vars vals
                        (handler-bind ((warning #'muffle-warning))
                          (apply function args))))
                    #-abcl
                    (handler-bind ((warning #'muffle-warning))
                      (apply function args)))))
               (try-stage-1 (message function)
                 (ignore-errors
                  (call-maybe-verbosely (format nil "Trying to ~A" message) function))
                 (maybe-done-stage-1))
               (try-file-stage-1 (explanation pathname)
                 (try-stage-1 (format nil "load ASDF from ~A" explanation)
                              #'(lambda () (load pathname))))
               (subpath (parent &key directory name type)
                 (merge-pathnames (make-pathname :defaults parent :version nil
                                                 :directory (cons :relative directory) :name name :type type)
                                  parent))
               (build/asdf.lisp (directory)
                 (subpath directory :directory '("build") :name "asdf" :type "lisp"))
               (visible-default-user-asdf-directory ()
                 (subpath (user-homedir-pathname) :directory '("common-lisp" "asdf")))
               (visible-default-user-asdf-lisp ()
                 (build/asdf.lisp (visible-default-user-asdf-directory)))
               (hidden-default-user-asdf-directory ()
                 (subpath (user-homedir-pathname) :directory '(".local" "share" "common-lisp" "asdf")))
               (hidden-default-user-asdf-lisp ()
                 (build/asdf.lisp (hidden-default-user-asdf-directory)))
               (stage-1 () ;; Try to load ASDF at all, any ASDF.
                 (try-stage-1
                  ;; Do nothing if ASDF is already loaded
                  "use an already loaded ASDF"
                  (constantly nil))
                 (try-stage-1
                  "require ASDF from the implementation"
                  ;; Most implementations provide ASDF, but while most of them are case-insensitive,
                  ;; CLISP is case-sensitive, so we need to specify a lowercase string,
                  ;; and not the keyword :asdf or symbol 'asdf.
                  ;; Most implementations provide ASDF 3, but some of them only ASDF 2
                  ;; and antique versions only ASDF 1.
                  #'(lambda () (funcall 'require "asdf")))
                 (try-file-stage-1
                  "asdf/ under the default (visible) CL source directory ~/common-lisp/"
                  (visible-default-user-asdf-lisp))
                 (try-file-stage-1
                  "asdf/ under the default (hidden) CL source directory ~/.local/share/common-lisp/"
                  (hidden-default-user-asdf-lisp))
                 #+(or unix linux bsd)
                 (progn
                   (try-file-stage-1
                    "asdf/ under the local system CL source directory /usr/local/share/common-lisp/"
                    (build/asdf.lisp #p"/usr/local/share/common-lisp/asdf/"))
                   (try-file-stage-1
                    "asdf/ under the managed system CL source directory /usr/share/common-lisp/"
                    (build/asdf.lisp #p"/usr/share/common-lisp/asdf/"))
                   (try-file-stage-1
                    "cl-asdf/ under the managed system CL source directory /usr/share/common-lisp/"
                    (build/asdf.lisp #p"/usr/share/common-lisp/cl-asdf/")))
                 (error "Could not load ASDF."))
               (maybe-done-stage-1 ()
                 ;; If we have ASDF, then go to stage 2: have it upgrade itself.
                 (when (member :asdf *features*)
                   (maybe-display (format nil "Found ASDF ~A" (asdf-version)))
                   (stage-2))) ;; doesn't return.
               (configure-asdf ()
                 ;; configure older versions of ASDF, as needed
                 (cond
                   ((probe-file (visible-default-user-asdf-lisp))
                    (unless (member :asdf3.1 *features*)
                      (maybe-display "Telling this old ASDF about your ~/common-lisp/asdf/")
                      (pushnew (visible-default-user-asdf-directory)
                               (symbol-value (asdf-symbol '*central-registry*)))))
                   ((probe-file (hidden-default-user-asdf-lisp))
                    (unless (member :asdf2 *features*)
                      (maybe-display "Telling this antique ASDF about your ~/.local/share/common-lisp/asdf/")
                      (pushnew (hidden-default-user-asdf-directory)
                               (symbol-value (asdf-symbol '*central-registry*)))))))
               (maybe-done-stage-2 ()
                 (when (ignore-errors (asdf-call 'version-satisfies
                                                 (asdf-version) required-asdf-version))
                   (when verbose
                     (format t "~&Victory! We now have ASDF ~A~%" (asdf-version)))
                   (return)))
               (stage-2 ()
                 ;; We have ASDF, now have it upgrade itself.
                 (configure-asdf)
                 (when (asdf-call 'find-system :asdf nil)
                   (call-maybe-verbosely
                    "Trying to upgrade ASDF"
                    (asdf-symbol 'operate) (asdf-symbol 'load-op) :asdf) :verbose nil)
                 (maybe-done-stage-2)
                 (error "We could only load ASDF ~A but we need ASDF ~A"
                        (asdf-version) required-asdf-version)))
        (call-maybe-verbosely nil #'stage-1))))

  (in-package :asdf))
;; Because of ASDF upgrade punting, this ASDF package may be a new one.
(eval-when (:compile-toplevel :load-toplevel :execute)
  (unless (or #+asdf2 (asdf:version-satisfies (asdf:asdf-version) "3.0.1"))
    (error "cl-launch requires ASDF 3.0.1 or later")))
;;;; Create cl-launch with UIOP.
(progn
(uiop:define-package :cl-launch
  (:use :common-lisp :uiop :asdf)
  (:export #:compile-and-load-file))

(in-package :cl-launch))
;;;; cl-launch initialization code
(progn
(defvar *cl-launch-file* nil) ;; name of this very file
(defvar *verbose* nil)
(progn
  (defun dump-stream-to-file (i n)
    (with-output-file (o n) (copy-stream-to-stream i o)))
  (defun dump-sexp-to-file (x n)
    (with-output-file (o n) (write x :stream o :pretty t :readably t)))
  (defvar *temporary-filenames* nil)
  (defvar *temporary-file-prefix*
    (format nil "~Acl-launch-~A-" *temporary-directory* (getenvp "CL_LAUNCH_PID")))
  (defun make-temporary-filename (x)
    (strcat *temporary-file-prefix* x))
  (defun register-temporary-filename (n)
    (push n *temporary-filenames*)
    n)
  (defun temporary-filename (x)
    (register-temporary-filename (make-temporary-filename x)))
  (defun temporary-file-from-foo (dumper arg x)
    (let ((n (temporary-filename x)))
      (funcall dumper arg n)
      n))
  (defun temporary-file-from-stream (i x)
    (temporary-file-from-foo #'dump-stream-to-file i x))
  (defun temporary-file-from-string (i x)
    (temporary-file-from-foo
     #'(lambda (i n) (with-output-file (o n) (princ i o))) i x))
  (defun temporary-file-from-sexp (i x)
    (temporary-file-from-foo #'dump-sexp-to-file i x))
  (defun temporary-file-from-file (f x)
    (with-open-file (i f :direction :input :if-does-not-exist :error)
      (temporary-file-from-stream i x)))
  (defun ensure-lisp-file-name (x &optional (name "load.lisp"))
    (let ((p (pathname x)))
      (if (equal (pathname-type p) "lisp")
          p
          (temporary-file-from-file p name))))
  (defun ensure-lisp-file (x &optional (name "load.lisp"))
    (cond
      ((eq x t) (temporary-file-from-stream *standard-input* "load.lisp"))
      ((streamp x) (temporary-file-from-stream x "load.lisp"))
      ((eq x :self) (ensure-lisp-file-name *cl-launch-file* name))
      (t (ensure-lisp-file-name x name))))
  (defun cleanup-temporary-files ()
    (loop :for n = (pop *temporary-filenames*)
          :while n :do
          (ignore-errors (delete-file n)))))
(defun file-newer-p (new-file old-file)
  "Returns true if NEW-FILE is strictly newer than OLD-FILE."
  (> (file-write-date new-file) (file-write-date old-file)))
(defun compile-and-load-file (source &key force-recompile
                              (verbose *verbose*) (load t)
                              output-file)
  "compiles and load specified SOURCE file, if either required by keyword
argument FORCE-RECOMPILE, or not yet existing, or not up-to-date.
Keyword argument VERBOSE specifies whether to be verbose.
Returns two values: the fasl path, and T if the file was (re)compiled"

  ;; When in doubt, don't trust - recompile. Indeed, there are
  ;; edge cases cases when on the first time of compiling a simple
  ;; auto-generated file (e.g. from the automated test suite), the
  ;; fasl ends up being written to disk within the same second as the
  ;; source was produced, which cannot be distinguished from the
  ;; reverse case where the source code was produced in the same split
  ;; second as the previous version was done compiling. Could be
  ;; tricky if a big system needs be recompiled as a dependency on an
  ;; automatically generated file, but for cl-launch those
  ;; dependencies are not detected anyway (BAD). If/when they are, and
  ;; lacking better timestamps than the filesystem provides, you
  ;; should sleep after you generate your source code.
  (let* (#+gcl
         (maybe-delete
           (unless (equal (pathname-type source) "lisp")
             (let ((temp (make-temporary-filename (strcat (pathname-name source) ".lisp"))))
               (copy-file source temp)
               (setf source temp))))
         (truesource (truename source))
         (fasl (or output-file (compile-file-pathname* truesource)))
         (compiled-p
          (when (or force-recompile
                    (not (probe-file fasl))
                    (not (file-newer-p fasl source)))
            (ensure-directories-exist fasl)
            (multiple-value-bind (path warnings failures)
                (compile-file* truesource :output-file fasl)
              (declare (ignorable warnings failures))
              (unless (equal (truename fasl) (truename path))
                (error "cl-launch: file compiled to ~A, expected ~A" path fasl))
              (when failures
                (error "cl-launch: failures while compiling ~A" source)))
            t)))
    (when load
      (load* fasl :verbose verbose))
    #+gcl (delete-file-if-exists maybe-delete)
    (values fasl compiled-p)))
(defun load-file (source &key output-file)
  (declare (ignorable output-file))
  #-(or gcl (and ecl (not dlopen)))
  (compile-and-load-file source :verbose *verbose* :output-file output-file)
  #+gcl
  (let* ((pn (parse-namestring source))) ; when compiling, gcl 2.6 will always
    (if (pathname-type pn) ; add a type .lsp if type is missing, to avoid compilation
      (compile-and-load-file source :verbose *verbose* :output-file output-file)
      (load source :verbose *verbose*)))
  #+(and ecl (not dlopen))
  (load source :verbose *verbose*))

(defun compute-arguments ()
  (setf *cl-launch-file* (getenvp "CL_LAUNCH_FILE")
        *verbose* (when (getenvp "CL_LAUNCH_VERBOSE") t)))

;; We provide cl-launch, no need to go looking for it further!
(unless (fboundp 'asdf::register-preloaded-system)
  (eval
   '(progn
      (defvar asdf::*preloaded-systems* (make-hash-table :test 'equal))
      (defun asdf::sysdef-preloaded-system-search (requested)
        (let ((name (coerce-name requested)))
          (multiple-value-bind (keys foundp) (gethash name asdf::*preloaded-systems*)
            (when foundp
              (apply 'make-instance 'system :name name :source-file (getf keys :source-file) keys)))))
      (defun asdf::register-preloaded-system (system-name &rest keys)
        (setf (gethash (coerce-name system-name) asdf::*preloaded-systems*) keys))
      (asdf::appendf asdf:*system-definition-search-functions* '(asdf::sysdef-preloaded-system-search)))))

(asdf::register-preloaded-system "cl-launch")

(defun build-and-load (build restart final init quit)
  (dolist (x build)
    (ecase (first x)
      ((:eval-input :load-system require)
       (symbol-call :asdf (first x) (second x)))
      ((:load)
       (let ((*package* (find-package (third x))))
         (etypecase (second x)
           (null nil)
           ((eql t) (load* nil))
           (stream (load* (second x)))
           ((eql :self) (load-file *cl-launch-file*))
           ((or pathname string) (load-file (second x))))))))
  (when final
    (eval-input final))
  (setf *image-prelude* init
        *image-entry-point* (when restart (ensure-function (car restart) :package (cdr restart)))
        *lisp-interaction* (not quit)))

(defun build-and-run (build restart final init quit)
  (build-and-load build restart final init quit)
  (restore-image))

#-ecl
(defun build-and-dump (dump build restart final init quit)
  (build-and-load build restart final init quit)
  (setf *features* (remove :cl-launched *features*))
  (dump-image dump :executable (getenvp "CL_LAUNCH_STANDALONE"))
  (quit 0))

;;; Attempt at compiling directly with ECL's make-build and temporary wrapper asd's
#+ecl
(progn
  (defvar *in-compile* nil)
  (defvar *dependency-counter* 0)
  (defun make-temporary-system (name options)
    (let* ((sys (pathname-name (temporary-filename name)))
           (asd (temporary-file-from-sexp `(defsystem ,sys ,@options) (strcat name ".asd"))))
      (asdf/find-system:load-asd asd :name sys)
      sys))
  (defclass asdf::cl-source-file-in-package (cl-source-file)
    ((package :initarg :package :reader component-package)))
  (defmethod perform :around ((o compile-op) (c asdf::cl-source-file-in-package))
    (let ((*package* (find-package (component-package c))))
      (call-next-method)))
  (defun make-dependency (fun arg pkg previous)
    (ecase fun
      ((:load)
       (let* ((load-file (ensure-lisp-file arg "load.lisp"))
              (dep-name (format nil "build-~D" *dependency-counter*)))
         (incf *dependency-counter*)
         (make-temporary-system
	  dep-name `(:depends-on ,previous
		     :components ((:cl-source-file-in-package ,(pathname-name load-file)
                                   :package ,pkg
				   :pathname ,(truename load-file)))))))
      ((:eval-input)
       (with-input (i arg)
         (make-dependency 'load i :cl-user previous)))
      ((:require)
       `(:require ,arg))
      ((:load-system)
       arg)))
  (defun build-and-dump (dump build restart final init quit)
    (unwind-protect
         (let* ((*compile-verbose* *verbose*)
                (*in-compile* t)
                (c::*suppress-compiler-warnings* (not *verbose*))
                (c::*suppress-compiler-notes* (not *verbose*))
                (*features* (remove :cl-launch *features*))
                (header (or *compile-file-pathname* *load-pathname* (getenvp "CL_LAUNCH_HEADER")))
                (header-file (ensure-lisp-file header "header.lisp"))
                (standalone (and (getenvp "CL_LAUNCH_STANDALONE") t))
                (footer
                  `(progn
		     (defun epilogue ()
                       (setf
                        *package* (find-package :cl-user)
                        *image-dumped-p* ,(if standalone :executable t)
                        *image-entry-point*
                        ,(when restart `(ensure-function ,(car restart) :package ,(cdr restart)))
                        *image-prelude* ,init
                        *image-postlude* ,final
                        *lisp-interaction* ,(not quit))
		       ,(if standalone
			    '(shell-boolean-exit (restore-image))
			    '(progn (si::top-level t) (quit))))
		     (unless *in-compile* (epilogue))))
		(footer-file (temporary-file-from-sexp footer "footer.lisp"))
                (dependencies
                  (loop :with r = ()
                        :for (fun arg pkg) :in
                        `((:load ,header-file :cl-user) ,@build (:load ,footer-file :cl-user))
                        :for dep = (make-dependency fun arg pkg r)
                        :do (setf r (append r (list dep)))
                        :finally (return r)))
                (program-sys
		  (make-temporary-system
		   "program"
		   `(:serial t
		     :build-operation program-op
		     :build-pathname ,(ensure-absolute-pathname dump #'getcwd)
		     :depends-on ,dependencies))))
           (operate 'program-op program-sys))
      (cleanup-temporary-files))
    (quit)))

(defun load-quicklisp ()
  (block nil
    (flet ((try (x) (when (probe-file* x) (return (load* x)))))
      (try (subpathname (user-homedir-pathname) ".quicklisp/setup.lisp"))
      (try (subpathname (user-homedir-pathname) "quicklisp/setup.lisp"))
      (error "Couldn't find quicklisp in your home directory. ~
              Go get it at http://www.quicklisp.org/beta/index.html"))))

(defun run (&key quicklisp source-registry build dump restart final init (quit 0))
  (pushnew :cl-launched *features*)
  (compute-arguments)
  (when source-registry (initialize-source-registry source-registry))
  (when quicklisp (load-quicklisp))
  (if dump
      (build-and-dump dump build restart final init quit)
      (build-and-run build restart final init quit)))

(pushnew :cl-launch *features*))

(cl-launch::compute-arguments)

;;; END OF CL-LAUNCH LISP HEADER
